# workers.py

import requests

from multiprocessing import Pool as ProcessPool
from multiprocessing.pool import ThreadPool


MSG = "\n\t{1.__class__.__name__}: {1.message} raised by {0.__name__} on {2}\n"


class Worker(object):
    """
    Wraps function which does work. When called, uses wrapped function on each
        input and outputs into buffer given.

    If wrapped function raises an exception, that item is excluded from output
        and the item is collected in error_buffer under self.error_key
        and the exception is collected under self.error_key + "_exception".

    If quiet flag is False, raised exceptions will throw an exception.

    When done ALL work, sets output_list.nomore to True.

    recip = Worker(lambda x: 1.0 / x)
    work, output, error = [2, 1, 0], [], {}
    recip(work, output, error)

    # wait a bit

    >>> output
    [.5, 1.0]
    """
    PRESERVES_ORDER = True
    UNPACK = False
    INSTANCES = 1
    QUIET = True

    def __init__(self, fn, unpack=None, instances=None, quiet=None,
                 error_key=None):
        self.fn = fn
        self.error_key = "{1.__name__} <{0.__class__.__name__}>".format(
            self, fn
        ) if error_key is None else error_key
        self.unpack = self.__class__.UNPACK if unpack is None else unpack
        self.instances = self.__class__.INSTANCES if instances is None else instances
        self.quiet = self.__class__.QUIET if quiet is None else quiet

    def push_to(self, output_list):
        def do(task):
            try:
                res = self.fn(task)
            except Exception as e:
                self.log(task, e)
                return None
            else:
                self.done(res, output_list)
        return do

    def process(self, work_list, output_list):
        """
        Default: single-threaded execution of work.
        Override this function for different workers.
        """
        map(self.push_to(output_list), work_list)

    def done(self, finished_product, output_list):
        if hasattr(finished_product, "__iter__") and self.unpack:
            output_list.extend(list(finished_product))
        else:
            output_list.append(finished_product)

    def log(self, item, exception):
        self.error_buffer[self.error_key].append({
            "item": item,
            "exception": exception
        })
        if not self.quiet:
            # don't raise exceptions here: tracebacks are misleading
            print MSG.format(self.fn, exception, item)

    def __call__(self, work_list, output_list, error_buffer):
        # Prepare for errors
        self.error_buffer = error_buffer  # for self.log function
        error_buffer[self.error_key] = error_buffer.get(self.error_key, [])

        # Delegate
        self.process(work_list, output_list)

        # Tell the next Trickler that no more is coming
        output_list.nomore = True


class ThreadWorker(Worker):
    """
    Preserves order. If one request takes too long, the worker will start on
        later work but will hold up the stream until the slow request finishes.
    """
    PRESERVES_ORDER = True
    INSTANCES = 6

    def process(self, work_list, output_list):

        def do(task):
            try:
                return self.fn(task)
            except Exception as e:
                self.log(task, e)
                return e
        # do = self.push_to(output_list)  # magically doesn't work

        pool = ThreadPool(self.instances)
        for result in pool.imap(do, work_list):
            if not isinstance(result, Exception):
                self.done(result, output_list)
        pool.close()
        pool.join()


class IOWorker(Worker):
    """
    Does not preserve order, in exchange for less bottlenecking downstream.
    """
    PRESERVES_ORDER = False
    INSTANCES = 12

    def process(self, work_list, output_list):
        pool = ThreadPool(self.instances)
        pool.map(self.push_to(output_list), work_list)
        pool.close()
        pool.join()


class ProcessWorker(Worker):
    PRESERVES_ORDER = False
    INSTANCES = 4

    def process(self, work_list, output_list):
        """
        NOTE: This is not preserving order right now.
        NOTE: This also may not work.
        """
        def do(task):
            try:
                res = self.fn(task)
            except Exception as e:
                self.log(task, e)
            finally:
                self.done(res, output_list)

        pool = ProcessPool(self.instances)
        pool.map(do, work_list)
        pool.close()
        pool.join()


def query(engine, sql, **defaults):
    def run_query(**kwargs):
        defaults.update(kwargs)
        sql = sql.format(**defaults)
        with engine.connection() as c:
            return c.execute(sql)
    return run_query


class API(object):
    def __init__(self, url):
        self.url = url

    @IOWorker
    def post(self, data):
        return requests.post(self.url, data=data)

    @IOWorker
    def get(self, params):
        return requests.get(self.url, params=params)




from fulforddata import constants as c
from fulforddata.access.accessors import access


def retrieve(data, form, *args):
    """
    Using accessors specified in form,
    retrieves all values from dictionary 'data'.
    form = {
        "value": accessor_function,
    }

    If accessor is a dict, recurses.
    If accessor is a str/unicode, splits on "/"

    Returns dictionary if succeeded. If accessor raises Exception,
    will return None (should be filtered out)
    and prints and logs a message about that entry.
    """
    entry = {}
    for tupl in form.items():
        key, accessor = tuple(tupl)  # just in case it's a list

        #
        # if accessor is a dict, recurse
        #
        if isinstance(accessor, c.DICTIONARIES):
            entry[key] = retrieve(data, accessor, *args)
            if entry[key] is None:  # retrieve failed
                return None  # filter it out
        #
        # else, access the value
        #
        else:
            #
            # if value is a string, use standard accessor with / as lister
            #
            if isinstance(accessor, c.STRINGS):
                accessor = access(accessor.split("/"))

            #
            # try to access the value from the data
            #
            try:
                entry[key] = accessor(data, *args)
            except Exception as e:
                #
                # accessing failed. Log message
                # and return None so it gets filtered out.
                #
                msg = "({}, {}): {}\nSkipping {}.".format(
                    *(key, accessor.__name__, str(e), data)
                )
                # logger.debug(msg)
                print msg
                return None
    return entry



# tools.py

from fulforddata import constants as c


def partition(fn, ls):
    """
    Calls fn on each item in ls
        returns dictionary keyed on outputs of fn
        values being lists of items with that output

    >>> partition(lambda x: x ** 2, [1, 2, 3, -2])
    {1: [1], 4: [2, -2], 9: [3]}

    >>> partition(bool, [1, 0, [], "Hello"])
    {True: [1, "Hello"], False: [0, []]}
    """
    parting = {}
    for item in ls:
        r = fn(item)
        try:
            parting[r].append(item)
        except KeyError:
            parting[r] = [item]
    return parting


def flatten_list(list_of_lists, depth=True):
    """Given a list of lists, unpacks all the items
    and puts it in one list.

    depth specifies how many layers deep we should flatten.
        by default flattens all layers

    >>> flatten_list([[1, [2], [3, 4]], [[5, [6, 7], [8]]]])
    [1, 2, 3, 4, 5, 6, 7, 8]

    >>> flatten_list([[1, [2], [3, 4]], [[5, [6, 7], [8]]]], depth=2)
    [1, 2, 3, 4, 5, [6, 7], [8]]
    """
    result = []
    for item in list_of_lists:
        if isinstance(item, c.ITERABLES):
            if type(depth) is int and depth - 1 >= 0:
                # flatten this level
                result.extend(flatten_list(item, depth=depth - 1))
            elif depth:
                # flatten all the way down
                result.extend(flatten_list(item))
            else:
                # preserve
                result.append(item)
        else:
            result.append(item)
    return result




import threading
from copy import deepcopy
from time import sleep


def mutate(key, mapper):
    def mutator(d):
        sleep(d[key])
        d[key] = mapper(d[key])
    return mutator


def fork(fns):

    def forker(item):
        tapestry = [threading.Thread(**{
            "target": f,
            "args": (item,)
        }) for f in fns]  # prepare worker threads
        [t.start() for t in tapestry]  # start worker threads
        [t.join() for t in tapestry]  # wait for all threads to finish
        return item

    return forker


if __name__ == "__main__":
    print fork([
        mutate("a", lambda x: x * 2),
        mutate("b", lambda x: x * 3)
    ])({"a": 2, "b": 4})



from inspect import getargspec

from fulforddata import constants as c


def prepare_argument(fn, contexts):
    """
    Passes only the arguments fn expects
        from contexts dicitonary
    """
    args, varargs, varkwargs, defaults = getargspec(fn)
    arguments = map(lambda a: contexts.get(a, None), args)
    return arguments


class ReadableFunction(object):
    """
    Decorator for functions.

    Acts like the function, with an eval-able .name attribute.
    If given function returns a callable when called:
        wraps callable with ReadableFunction,
        and fills in the .name attribute to hold args used in call
    Else, just returns the value.

    __str__ or __repr__ will access .name.

    >>> @ReadableFunction
    ... def sample(a, b, c=42):
    ...     def inner_func():
    ...         return a + b + c
    ...     return inner_func

    >>> str(sample(1, 42, c=73))
    'sample(1, 42, c=73)'
    """
    def __init__(self, fn, name=None):
        self.fn = fn
        self.name = name if isinstance(name, c.STRINGS) else fn.__name__
        self.__name__ = self.name
        self.__doc__ = fn.__doc__ if hasattr(fn, "__doc__") else self.__doc__

    def __str__(self):
        return self.name

    __repr__ = __str__

    def _resolve_to_py_function(self):
        """
        So the inspect module can work, call this to get the
            actually defined python function
        """
        if isinstance(self.fn, ReadableFunction):
            return self.fn._resolve_to_py_function()
        else:
            return self.fn

    def __call__(self, *args, **kwargs):
        result = self.fn(*args, **kwargs)

        # if a function is returned,
        # make it a validator
        # with the name describing how it was made.
        if hasattr(result, "__call__"):
            name = self.name + "("
            if args or kwargs:
                arguments = []
                if args:
                    typetype = type(type(1))
                    args = map(lambda x: x.__name__ if isinstance(x, typetype)
                               else repr(x), args)
                    arguments.extend(args)
                if kwargs:
                    arguments.extend(map(lambda kv: str(kv[0]) + "=" +
                                         repr(kv[1]), kwargs.items()))
                name += ", ".join(arguments)
            name += ")"
            # print name
            return ReadableFunction(result, name=name)
        return result


def compose(*fns):
    """
    Returns a function that evaluates each function in order, starting with
        function 0's output being fed into function 1's input.

    The first function can accept anything, but output beyond that is passed
        as a single value.
    """
    def comp(*args, **kwargs):
        v = fns[0](*args, **kwargs)
        for fn in fns[1:]:
            v = fn(v)
        return v
    return comp


def pass_through(value, mapping, warn=False, **kwargs):
    """Tries to map value through mapping. If fails, returns input.
    Mapping can be a function or a dictionary.

    If function, key words are passed down to the function.
        Exceptions are caught and printed, returns value

    If dictionary, tries to get new value by inputting the value.
        If value is not hashable (list, set, etc.), prints and returns value

    >>> pass_through(2, lambda x: x ** 2)
    4

    >>> pass_through("color", {"color": "BLUE"})
    "BLUE"

    >>> pass_through("color", {"id": 42})
    "color"

    >>> pass_through(["unhashable"], {"id": 42})
    ["unhashable"]  # also prints error to stdout

    >>> pass_through(0, lambda x: 1/x)
    0  # prints error to stdout

    """

    # Lists cannot pass_through. Instead, we pass through all items.
    # If a staggered, unflattened list of lists,
    #   will pass all entries and lower entries through
    if isinstance(value, c.ITERABLES):
        return map(lambda v: pass_through(v, mapping, warn=False, **kwargs),
                   value)

    if isinstance(mapping, c.DICTIONARIES):
        try:
            return mapping.get(value, value)
        except TypeError:
            msg = "pass_through: TypeError (unhashable type?)" \
                ": cannot get {} from {}".format(value, mapping)
            print msg
            # will return value later
        except KeyError:
            # this is expected behavior
            pass  # will return value later
    else:
        try:
            return mapping(value, **kwargs)
        except Exception as e:  # calling the function failed
            msg = "pass_through: {}({}) failed\n".format(
                mapping, value) + str(e)
            print msg

    # Mapping failed - return input
    return value


if __name__ == "__main__":
    print pass_through(1, {1: 9})




from fulforddata import constants as c
from fulforddata.utilities.functional import ReadableFunction, prepare_argument


def check(data, template, **context):
    """
    Checks data complies with tests specified in template.
    If test fails, prints value returned and returns False
        continues to evaluate tests
    Is part of validation.
    """

    if isinstance(template, c.DICTIONARIES):
        # delegate to all key-value pairs
        results = []
        for key, value in template.items():
            result = check(data[key], value, **context)
            results.append(result)
            if not result:
                break
        return reduce(lambda p, q: bool(p) and bool(q), results)
    elif isinstance(template, c.ITERABLES):
        # delegate to all items
        results = []
        for item in data:  # data should be a list
            result = check(item, template[0], **context)
            results.append(result)
            if not result:
                break
        return reduce(lambda p, q: bool(p) and bool(q), results)
    else:
        # Leaf of template
        contexts = {"data": data}
        contexts.update(context)

        test = template
        if isinstance(test, ReadableFunction):
            test = test._resolve_to_py_function()
        try:
            # Call the validator on this piece of data
            arguments = prepare_argument(test, contexts)
            output = test(*arguments)
        except Exception as e:
            output = str(e)
            raise

        if output is not True:
            print "{}({}) failed (output: {})" \
                .format(template.__name__, data, output)
            return False
        return True



# stream.py

import time
import threading
from copy import deepcopy

import workers


class Trickle(object):

    def __init__(self, work):
        self.work = work
        self.i = 0
        self.nomore = False

    def __iter__(self):
        return self

    def __getitem__(self, i):
        return self.work[i]

    def __repr__(self):
        return "Tricker({})".format(self.work)

    __str__ = __repr__

    def append(self, item):
        self.work.append(item)

    def extend(self, items):
        self.work.extend(items)

    def next(self):
        try:
            ret = self.work[deepcopy(self.i)]
            self.i += 1
            return ret if ret is not None else self.next()
        except KeyError:
            # print "{} out of work. Nomore: {}".format(self.i, self.nomore)
            if self.nomore:
                raise StopIteration()
            time.sleep(0.05)
            return self.next()  # TODO: make iterative; limit depth
        except IndexError:
            # print "{} out of work. Nomore: {}".format(self.i, self.nomore)
            if self.nomore:
                raise StopIteration()
            time.sleep(0.05)
            return self.next()  # TODO: make iterative; limit depth


class Stream(object):
    """
    Represents a task that can be streamlined. Useful for doing work while
        waiting for IO.

    NOTE: May not preserve order, depending on which workers are used.
        Use .preserves_order attr to find if your Stream should preserve order.

    >>> mystream = Stream().then(
        lambda x: x ** 2,
        name="Square"
    ).then(  # [0, 1, 4, 9]
        lambda x: [i for i in range(x)],
        unpack=True,
        name=
    )

    >>> mystream(range(4))
    [0, 1, 0, 1, 2, 3, 0, 1, 2, 3, 4, 5, 6, 7, 8]

    >>> mystream.errors
    {}

    >>> mystream.preserves_order
    True

    >>> mystream.then(
        lambda x: 1.0 / x,
        name="Reciprocal"
    )
    mystream

    >>> mystream(range(4))
    [1, 1, 2, 3, 1, 2, 3, 4, 5, 6, 7, 8]

    >>> mystream.errors
    {
        "Reciprocal <Worker>": [
            {"item": 0, "exception": DivideByZeroException},  # 0th index
            {"item": 0, "exception": DivideByZeroException},  # 2nd index
            {"item": 0, "exception": DivideByZeroException},  # 6th index
        ]
    }
    """
    def __init__(self, default_worker=workers.Worker, **worker_args):
        self._workers = []
        self.errors = {}
        self.default_worker = default_worker
        self.defaults = worker_args

    @property
    def preserves_order(self):
        return all(w.PRESERVES_ORDER for w in self._workers)

    def then(self, worker, **worker_overrides):
        """
        Appends this worker (or these workers) at the end of the stream.

        If worker variable is not already a worker, wraps with this stream's
            default worker (as specified during __init__)

        Returns this stream.
        """
        if hasattr(worker, "__iter__"):
            map(self.then, worker)
        else:
            if not isinstance(worker, workers.Worker):
                worker_args = deepcopy(self.defaults)
                worker_args.update(worker_overrides)
                worker = self.default_worker(worker, **worker_args)
            worker.error_key = len(self._workers)
            self._workers.append(worker)
        return self

    def __call__(self, work):
        work = deepcopy(work)  # in case initial input is important

        work_queues = [work] + [[] for w in self._workers]
        work_queues = map(Trickle, work_queues)
        work_queues[0].nomore = True  # no more input in Tricker 0

        # print "Before: {}".format(work_queues)

        tapestry = [threading.Thread(**{
            "target": self._workers[i],
            "args": (work_queues[i], work_queues[i + 1], self.errors)
        }) for i in range(len(self._workers))]  # prepare worker threads
        [t.start() for t in tapestry]  # start worker threads

        return work_queues[-1]
        # [t.join() for t in tapestry]  # wait for all threads to finish

        # print "After: {}".format(work_queues)

        # return work_queues[-1].work  # todo: yield as results trickle in


if __name__ == "__main__":
    def wait(d):
        val = d["value"]
        if val > 5:
            raise Exception("TOO LONG")
        time.sleep(val)
        return d

    def recip(d):
        d["value"] = 1.0 / d["value"]
        return d

    def tee(d):
        print d
        return d

    #
    # Example
    #

    mystream = Stream(quiet=False).then([
        lambda x: {"value": x},
        workers.IOWorker(wait),
        tee,
    ])

    results = mystream([2, 3, 0, 1, 7])
    results.extend(mystream([4, 4, 9]))
    for i in results:
        print "Done:", i

    print
    print "Errors"
    for e in sorted(mystream.errors.items()):
        print e

    print "This stream does {}preserve order.".format(
        "" if mystream.preserves_order else "NOT "
    )
